
import pytest

import dragodis
from dragodis import LineType, CommentType, interface


@pytest.mark.parametrize("address,line_type,value,size,data", [
    (0x401003, LineType.code, None, 3, b"\x8b\x45\x08"),
    (0x40c0c4, LineType.undefined, 0x5c, 1, b"\x5c"),
    (0x40ec38, LineType.unloaded, None, 1, b""),
    (0x40c000, LineType.string, "Idmmn!Vnsme ", 13, b"Idmmn!Vnsme \x00"),
    (0x40b8bc, LineType.word, 0x186, 2, b"\x86\x01"),
    (0x40b8be, LineType.string, "GetCommandLineA", 16, b"GetCommandLineA\x00"),
    (0x40b784, LineType.dword, 0xb7ac, 4, b"\xac\xb7\x00\x00"),
    (0x40a838, LineType.string16, "KERNEL32.DLL", 26,
        b"K\x00E\x00R\x00N\x00E\x00L\x003\x002\x00.\x00D\x00L\x00L\x00\x00\x00"),
])
def test_basic(disassembler, address, line_type, value, size, data):
    """
    Basic test for getting line type, value, size, and data
    """
    line = disassembler.get_line(address)
    assert line.address == address
    assert line.type == line_type
    assert line.size == size
    assert line.data == data
    if line_type == LineType.code:
        assert isinstance(line.value, interface.Instruction)
    else:
        assert line.value == value


def test_align_type_and_value(disassembler):
    """
    For alignment bytes, the disassembler may mark it as either LineType.undefined
    or LineType.align depending on the disassembler and analysis options.
    Therefore, for this special case test both possibilities.

    For IDA, alignment bytes are detected by default.
    For Ghidra, alignment bytes are not detected by default.
        The "Condense Filler Bytes" analysis options must be enabled.
    """
    line = disassembler.get_line(0x40102b)
    assert line.type in (LineType.align, LineType.undefined)

    if line.type == LineType.undefined:
        assert line.size == 1
        assert line.value == 0xCC
        # If not already alignment, test setting alignment to prove
        # the disassembler can properly handle alignment bytes.
        line.type = LineType.align

    assert line.size == 5
    assert line.data == b"\xCC" * 5
    assert line.value == b"\xCC" * 5


@pytest.mark.parametrize("address,new_type,value", [
    (0x40c0f0, LineType.undefined, 0x43),
    # FIXME: Inconsistencies in precision between disassemblers.
    #   Determine appropriate standard for handling floats.
    pytest.param(0x40c0f0, LineType.float, 7.157251582566349e+22, marks=pytest.mark.xfail),
    (0x40c0f0, LineType.dword, 0x65727f43),
    (0x40c0f0, LineType.word, 0x7f43),
    # NOTE: Disassembler will end up pulling a different amount of bytes based on disassembler.
    (0x40c000, LineType.string, "Idmmn!Vnsme "),
])
def test_setting_type(disassembler, address, new_type, value):
    """
    Tests setting line to a new type and it's effects on the value.
    """
    line = disassembler.get_line(address)
    orig_value = line.value
    orig_type = line.type

    # Test setting a new type and how it affects the value.
    line.type = new_type
    assert line.type == new_type
    assert line.value == value

    # Reset
    line.type = orig_type
    assert line.type == orig_type
    assert line.value == orig_value


@pytest.mark.parametrize("address,data,value", [
    # (0x401003, None),    # TODO: Add setting new instruction
    (0x40c0c4, b"\xff", 0xff),  # undefined
    (0x40ec38, b"", None),  # unloaded
    (0x40c000, b"hello\x00", "hello"),  # string
    (0x40b8bc, b"\x01\x00", 0x1),  # word
    (0x40b8be, b"hello\x00", "hello"),  # string
    (0x40b784, b"\xef\xcd\xab\x00", 0xabcdef),  # dword
    (0x40a838, b"h\x00e\x00l\x00l\x00o\x00\x00\x00", "hello"),  # string16
])
def test_setting_data_and_value(disassembler, address, data, value):
    """
    Tests setting line to a new value or data and its affect on the other.
    Also insures the line type doesn't get changed.
    """
    line = disassembler.get_line(address)
    orig_value = line.value
    orig_data = line.data
    orig_type = line.type

    # Test setting the value
    line.value = value
    assert line.value == value
    assert line.data == data
    assert line.type == orig_type
    line.value = orig_value
    assert line.value == orig_value
    assert line.data == orig_data
    assert line.type == orig_type

    # Test setting the data
    line.data = data
    assert line.data == data
    assert line.value == value
    assert line.type == orig_type
    line.data = orig_data
    assert line.data == orig_data
    assert line.value == orig_value
    assert line.type == orig_type


def test_setting_value_with_new_type(disassembler):
    """
    Tests that setting a value could trigger a change in type
    based on the type of value set.
    """
    line = disassembler.get_line(0x40c000)
    orig_value = line.value
    orig_data = line.data

    # Originally a string
    assert line.type == LineType.string
    assert isinstance(orig_value, str)

    # Setting value with an int, should change it to a dword.
    line.value = 0xabcd
    assert line.type == LineType.dword
    assert line.value == 0xabcd
    assert line.data == b"\xcd\xab\x00\x00"

    # Setting value as a float, should change it to a float.
    line.value = 4.5
    assert line.type == LineType.float
    assert line.value == 4.5

    # Setting value as single byte sets it to byte type.
    line.value = b"\x01"
    assert line.type == LineType.byte
    assert line.value == 0x1
    assert line.data == b"\x01"

    # Setting back to original string, should put it back as a string.
    line.value = orig_value
    assert line.type == LineType.string
    assert line.value == orig_value
    assert line.data == orig_data


def test_name(disassembler):
    """
    Tests getting and seting a name on a line.
    """
    line = disassembler.get_line(0x40c0c4)
    orig_name = line.name
    assert orig_name
    line.name = "new_name"
    assert line.name == "new_name"
    line.name = None
    assert line.name == orig_name


def test_prev_next(disassembler):
    """
    Tests getting previous and next line.
    """
    # Test getting next line of instruction.
    line = disassembler.get_line(0x401035)
    assert line.prev.address == 0x401033
    assert line.next.address == 0x40103a

    # Test getting next line of undefined data.
    line = disassembler.get_line(0x40c0c4)
    assert line.prev.address == 0x40c0c3
    # TODO: Look into standardizing this.
    # IDA treats all undefined data in chunks.
    if disassembler.name.lower() == "ida":
        assert line.next.address == 0x40c114  # Jumps to next defined item.
    # Ghidra treats each individual byte of undefined data as a separate line.
    elif disassembler.name.lower() == "ghidra":
        assert line.next.address == 0x40c0c5  # Just the next byte
    else:
        pytest.fail(f"Update test for disassembler: {disassembler.name}")


def test_instruction(disassembler):
    line = disassembler.get_line(0x401035)
    assert line.instruction
    assert isinstance(line.instruction, interface.Instruction)
    line = disassembler.get_line(0x40c0c4)
    assert line.instruction is None


@pytest.mark.parametrize("address", [
    0x401035,  # code address
    0x40c130,  # data address
    0x40c14f,  # undefined address
])
@pytest.mark.parametrize("comment,comment_type", [
    ("eol comment", CommentType.eol),
    ("This is an\nanterior comment", CommentType.anterior),
    ("This is a\nposterior comment", CommentType.posterior),
    ("This is a\nplate comment", CommentType.plate),
    ("repeatable comment", CommentType.repeatable),
], ids=["eol", "anterior", "posterior", "plate", "repeatable"])
def test_comment(disassembler, address, comment, comment_type):
    """
    Tests getting and setting comments
    NOTE: This is more or less a copy of the test in test_flat.py
    """
    # Test setting comment
    line = disassembler.get_line(address)
    assert line.get_comment(comment_type=comment_type) is None
    line.set_comment(comment, comment_type=comment_type)
    assert line.get_comment(comment_type=comment_type) == comment
    # Test resetting comment
    line.set_comment(None, comment_type=comment_type)
    assert line.get_comment(comment_type=comment_type) is None




import pytest

from dragodis import FlowType


@pytest.mark.parametrize("address,block_list", [
    (0x004011AA, [
        (0x00401150, 0x004012A0, FlowType.terminal),
    ]),
    (0x004035BB, [
        (0x00403597, 0x004035AB, FlowType.conditional_jump),
        (0x004035AB, 0x004035B1, FlowType.conditional_jump),
        (0x004035B1, 0x004035B3, FlowType.call),
        (0x004035B3, 0x004035BA, FlowType.conditional_jump),
        (0x004035BA, 0x004035BD, FlowType.terminal),
    ]),
    (0x00401000, [
        (0x00401000, 0x00401003, FlowType.fall_through),
        (0x00401003, 0x0040100D, FlowType.conditional_jump),
        (0x0040100D, 0x00401029, FlowType.unconditional_jump),
        (0x00401029, 0x0040102B, FlowType.terminal),
    ]),
])
def test_blocks_start_end_flow_type(disassembler, address, block_list):
    flowchart = disassembler.get_flowchart(address)
    assert flowchart
    assert len(flowchart) == len(block_list)
    assert [(block.start, block.end, block.flow_type) for block in flowchart.blocks] == block_list


@pytest.mark.parametrize("address,blocks_to,blocks_from", [
    (0x401003, [0x401000, 0x40100D], [0x40100D, 0x401029]),
    (0x401029, [0x401003], []),
    (0x4035B3, [0x4035AB, 0x4035B1], [0x4035AB, 0x4035BA]),
    pytest.param(
        0x402cfe, [0x402cce, 0x402cf0], [0x402d3a],
        marks=pytest.mark.xfail(reason="Ghidra fails to analyze _invoke_watson as a no return function.")
    ),
])
def test_blocks_to_from(disassembler, address, blocks_to, blocks_from):
    block = disassembler.get_basic_block(address)
    assert block
    assert sorted([b.start for b in block.blocks_to]) == blocks_to
    assert sorted([b.start for b in block.blocks_from]) == blocks_from


def test_equality(disassembler):
    flowchart = disassembler.get_flowchart(0x004035BB)
    flowchart2 = disassembler.get_flowchart(0x004035AB)
    assert flowchart and flowchart2
    assert flowchart == flowchart2


def test_get_block(disassembler):
    flowchart = disassembler.get_flowchart(0x004035BB)
    assert flowchart
    found_block = flowchart.get_block(0x004035AD)
    assert found_block
    assert found_block.start == 0x004035AB


@pytest.mark.parametrize("address,lines", [
    (0x00401003, [
        0x00401003, 0x00401006, 0x00401009, 0x0040100B,
    ]),
    (0x004011AA, [
        0x00401150, 0x00401151, 0x00401153, 0x00401158, 0x0040115D, 0x00401162, 0x00401167,
        0x0040116A, 0x0040116F, 0x00401174, 0x00401179, 0x0040117C, 0x00401181, 0x00401186,
        0x0040118B, 0x0040118E, 0x00401193, 0x00401198, 0x0040119D, 0x004011A0, 0x004011A5,
        0x004011AA, 0x004011AF, 0x004011B2, 0x004011B7, 0x004011BC, 0x004011C1, 0x004011C4,
        0x004011C9, 0x004011CE, 0x004011D3, 0x004011D6, 0x004011DB, 0x004011E0, 0x004011E5,
        0x004011E8, 0x004011ED, 0x004011F2, 0x004011F7, 0x004011FA, 0x004011FF, 0x00401204,
        0x00401209, 0x0040120C, 0x00401211, 0x00401216, 0x0040121B, 0x0040121E, 0x00401223,
        0x00401228, 0x0040122D, 0x00401230, 0x00401235, 0x0040123A, 0x0040123F, 0x00401242,
        0x00401247, 0x0040124C, 0x00401251, 0x00401254, 0x00401259, 0x0040125E, 0x00401263,
        0x00401266, 0x0040126B, 0x00401270, 0x00401275, 0x00401278, 0x0040127D, 0x00401282,
        0x00401287, 0x0040128A, 0x0040128F, 0x00401294, 0x00401299, 0x0040129C, 0x0040129E,
        0x0040129F,
    ]),
])
def test_block_lines(disassembler, address, lines):
    block = disassembler.get_basic_block(address)
    assert block
    assert [line.address for line in block.lines()] == lines


def test_flowchart_from_block(disassembler):
    flowchart = disassembler.get_flowchart(0x004011AA)
    block = disassembler.get_basic_block(0x004011AA)
    assert block.flowchart == flowchart


@pytest.mark.parametrize("address,ancestors", [
    # first block
    (0x403597, []),
    # block in the middle with a loop
    (0x4035B1, [
        0x403597,
        0x4035ab,
        0x4035b3,  # loop back
    ]),
    # last block
    (0x4035BB, [
        0x403597,
        0x4035ab,
        0x4035b1,
        0x4035b3,
    ])
])
def test_ancestors(disassembler, address, ancestors):
    flowchart = disassembler.get_flowchart(address)
    block = flowchart.get_block(address)
    assert sorted(b.start for b in block.ancestors) == ancestors


def test_traverse(disassembler):
    flowchart = disassembler.get_flowchart(0x004035BB)

    found_block = flowchart.get_block(0x004035AD)
    assert found_block
    assert found_block.start == 0x004035AB

    blocks = list(flowchart.traverse(start=0x004035AB, reverse=True))
    assert len(blocks) == 2
    assert [(block.start, block.end) for block in blocks] == [
        (0x004035AB, 0x004035B1),
        (0x00403597, 0x004035AB),
    ]

    blocks = list(flowchart.traverse(start=0x004035AB))
    assert len(blocks) == 4
    assert [(block.start, block.end) for block in blocks] == [
        (0x004035AB, 0x004035B1),
        (0x004035BA, 0x004035BD),
        (0x004035B1, 0x004035B3),
        (0x004035B3, 0x004035BA),
    ]

    blocks = list(flowchart.traverse())
    assert len(blocks) == 5
    assert [(block.start, block.end) for block in blocks] == [
        (0x00403597, 0x004035AB),
        (0x004035AB, 0x004035B1),
        (0x004035BA, 0x004035BD),
        (0x004035B1, 0x004035B3),
        (0x004035B3, 0x004035BA),
    ]
    blocks = list(flowchart.traverse(reverse=True))
    assert len(blocks) == 5
    assert [(block.start, block.end) for block in blocks] == [
        (0x004035BA, 0x004035BD),
        (0x004035B3, 0x004035BA),
        (0x00403597, 0x004035AB),
        (0x004035B1, 0x004035B3),
        (0x004035AB, 0x004035B1),
    ]
    blocks = list(flowchart.traverse(dfs=True))
    assert len(blocks) == 5
    assert [(block.start, block.end) for block in blocks] == [
        (0x00403597, 0x004035AB),
        (0x004035AB, 0x004035B1),
        (0x004035B1, 0x004035B3),
        (0x004035B3, 0x004035BA),
        (0x004035BA, 0x004035BD),
    ]
    blocks = list(flowchart.traverse(reverse=True, dfs=True))
    assert len(blocks) == 5
    assert [(block.start, block.end) for block in blocks] == [
        (0x004035BA, 0x004035BD),
        (0x004035B3, 0x004035BA),
        (0x004035B1, 0x004035B3),
        (0x004035AB, 0x004035B1),
        (0x00403597, 0x004035AB),
    ]



import pytest

import dragodis
from dragodis import CommentType
from dragodis import interface


def test_disassembler_info_ida_all(disassembler):
    assert disassembler.name == "IDA"


def test_disassembler_info_ghidra_all(disassembler):
    assert disassembler.name == "Ghidra"


def test_processor_info(disassembler):
    assert disassembler.bit_size == 32
    assert disassembler.processor_name == "x86"
    assert disassembler.compiler_name in ("Visual C++", "visualstudio:unknown")
    assert disassembler.is_big_endian == False


def test_processor_info_arm(disassembler):
    assert disassembler.bit_size == 32
    assert disassembler.processor_name == "ARM"
    assert disassembler.compiler_name in ("GNU C++", "unknown")  # TODO: Ghidra fails to figure this one out.
    assert disassembler.is_big_endian == False


def test_addressing_ida(disassembler):
    assert disassembler.min_address == 0x401000
    assert disassembler.max_address == 0x40f000


def test_addressing_ghidra(disassembler):
    assert disassembler.min_address == 0x400000
    # Ghidra 10.2.* includes a new "tdb" memory block which increases the max address size.
    # https://github.com/NationalSecurityAgency/ghidra/issues/4790
    assert disassembler.max_address in (0x40ed47, 0xffdfffff)


def test_entry_point_x86(disassembler):
    assert disassembler.entry_point == 0x4014e0


def test_entry_point_arm(disassembler):
    assert disassembler.entry_point == 0x1030c


def test_base_address_x86(disassembler):
    assert disassembler.base_address == 0x400000


def test_base_address_arm(disassembler):
    assert disassembler.base_address == 0x10000


def test_file_offset(disassembler):
    assert disassembler.get_file_offset(0x4063b7) == 0x57b7
    assert disassembler.get_virtual_address(0x57b7) == 0x4063b7
    assert disassembler.get_file_offset(0x40c000) == 0xae00
    assert disassembler.get_virtual_address(0xae00) == 0x40c000
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_file_offset(0x1)
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_virtual_address(0xc001)


def test_is_loaded(disassembler):
    assert disassembler.is_loaded(0x4063b7)
    assert not disassembler.is_loaded(0x117f7d8)


def test_lines(disassembler):
    # Test getting a line
    line = disassembler.get_line(0x40116A)
    assert line
    assert line.address == 0x40116A

    # Test code section
    expected = [
        0x401003, 0x401006, 0x401009, 0x40100B, 0x40100D, 0x401011, 0x401014,
        0x401017, 0x401019, 0x40101C, 0x40101E, 0x401021, 0x401024, 0x401027,
    ]
    assert [line.address for line in disassembler.lines(start=0x401003, end=0x401029)] == expected
    assert list(disassembler.line_addresses(start=0x401003, end=0x401029)) == expected

    # Test data section
    if disassembler.name.lower() == "ida":
        expected = [
            0x40c000,
            0x40c00d,
            0x40c010,
            0x40c02a,
            0x40c02c,
            0x40c059,
            0x40c05c,
            0x40c080,
            0x40c0a0,
            0x40c0c3,
        ]
    # Ghidra doesn't have the concept of alignment bytes and instead treats them as individual
    # undefined bytes.
    elif disassembler.name.lower() == "ghidra":
        expected = [
            0x40c000,
            *range(0x40c00d, 0x40c010),
            0x40c010,
            *range(0x40c02a, 0x40c02c),
            0x40c02c,
            *range(0x40c059, 0x40c05c),
            0x40c05c,
            0x40c080,
            0x40c0a0,
            *range(0x40c0c3, 0x40c0e8),
        ]

    assert [line.address for line in disassembler.lines(start=0x40C000, end=0x40C0E8)] == expected
    assert list(disassembler.line_addresses(start=0x40C000, end=0x40C0E8)) == expected


def test_references(disassembler):
    refs = list(disassembler.references_to(0x40c15c))
    assert len(refs) == 2
    assert all(ref.to_address == 0x40c15c for ref in refs)
    assert sorted(ref.from_address for ref in refs) == [0x4010f8, 0x401242]

    refs = list(disassembler.references_from(0x401242))
    assert len(refs) == 1
    assert refs[0].from_address == 0x401242
    assert refs[0].to_address == 0x40c15c


# # TODO: Add test for duplicate name
# # TODO: Figure out how to manage the resetting of data addresses.
# #   - when I set aTsudfs it will actually set to asc_40C130
# #   - Is there a way to control the way IDA names strings? (or get the original string name?)
# #       - (Remember, we can check the LineType!!!!)
#
# # TODO: Set this unit test to be generic from the dissassembler.
# @pytest.mark.parametrize(
#     "addr,expected",
#     [
#         # Function address
#         (0x401030, {"ida": "sub_401030", "ghidra": "FUN_00401030"}),
#         # Data address
#         (0x40c130, {"ida": ("aTsudfs", "asc_40C130"), "ghidra": "s_tSUdFS_0040c130"}),
#         # No original name (in the middle of unknown data)
#         (0x40c14f, {"ida": None, "ghidra": None}),
#     ]
# )
# def test_name(disassembler, addr, expected):
#     expected_names = expected[disassembler.name.lower()]
#     if not isinstance(expected_names, tuple):
#         expected_names = (expected_names,)
#
#     assert disassembler.get_name(addr) in expected_names
#     disassembler.set_name(addr, "test_name")
#     assert disassembler.get_name(addr) == "test_name"
#     # Setting name to None or empty string should reset it.
#     disassembler.set_name(addr, None)
#     assert disassembler.get_name(addr) in expected_names


@pytest.mark.parametrize("address,is_defined", [
    (0x401030, True),   # function address
    (0x40c130, True),   # data address
    (0x40c14f, False),  # undefined address
])
def test_name(disassembler, address, is_defined):
    # Different disassemblers are going to give different names, so we aren't
    # going to test what those names actually are.
    original_name = disassembler.get_name(address)
    if is_defined:
        assert original_name and isinstance(original_name, str)
    else:
        assert original_name is None

    disassembler.set_name(address, "test_name")
    assert disassembler.get_name(address) == "test_name"

    # Setting name to None or empty string should reset it.
    disassembler.set_name(address, None)
    assert disassembler.get_name(address) == original_name

    # Test duplicate name generation.

    # TODO: Test duplicate name generation.


@pytest.mark.parametrize("address", [
    0x401030,  # function address
    0x401035,  # code address
    0x40c130,  # data address
    0x40c14f,  # undefined address
])
@pytest.mark.parametrize("comment,comment_type", [
    ("eol comment", CommentType.eol),
    ("This is an\nanterior comment", CommentType.anterior),
    ("This is a\nposterior comment", CommentType.posterior),
    ("This is a\nplate comment", CommentType.plate),
    ("repeatable comment", CommentType.repeatable),
], ids=["eol", "anterior", "posterior", "plate", "repeatable"])
def test_comment(disassembler, address, comment, comment_type):
    # Test setting comment
    assert disassembler.get_comment(address, comment_type=comment_type) is None
    disassembler.set_comment(address, comment, comment_type=comment_type)
    assert disassembler.get_comment(address, comment_type=comment_type) == comment
    # Test resetting comment
    disassembler.set_comment(address, None, comment_type=comment_type)
    assert disassembler.get_comment(address, comment_type=comment_type) is None


def test_get_function(disassembler):
    # Test getting a Function object
    func = disassembler.get_function(0x401030)
    assert func
    assert isinstance(func, interface.Function)
    assert func.start == 0x401030

    # Test getting same function not at entrypoint
    func = disassembler.get_function(0x401035)
    assert func
    assert isinstance(func, interface.Function)
    assert func.start == 0x401030

    # Test invalid function
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_function(0x40c130)

    # Test getting multiple functions
    # Number of functions are different for each disassembler, but we should
    # get at least 200.
    funcs = list(disassembler.functions())
    assert len(funcs) > 200


def test_get_function_by_name(disassembler):
    func = disassembler.get_function_by_name("printf")
    assert func
    assert func.start == 0x4012a0
    assert "printf" in func.name


@pytest.mark.parametrize("backend", ["ida", "ghidra"])
def test_create_function(shared_datadir, backend):
    input_path = shared_datadir / "strings_x86 .text[00401000,0040102a].bin"
    try:
        with dragodis.open_program(str(input_path), backend, processor=dragodis.PROCESSOR_X86) as dis:
            if func := dis.get_function(0x0, None):
                func.undefine(True)
            assert not dis.get_function(0x0, None)
            func = dis.create_function(0x0)
            assert func
            assert func.start == 0
            assert func.end == 0x2B
    except dragodis.NotInstalledError as e:
        pytest.skip(str(e))


def test_functions(disassembler):
    # Test getting all functions
    funcs = list(disassembler.functions())
    # NOTE: Exact number of functions is highly dependent on the disassembler and
    # it's default analysis settings and could change in future versions.
    # Therefore, just test that we have decent amount of functions and some expected functions are
    # there.
    assert len(funcs) >= 200  # IDA had 220 and Ghidra had 226
    func_names = [func.name for func in funcs]
    func_addrs = [func.start for func in funcs]
    assert "_printf" in func_names
    assert 0x401030 in func_addrs

    # TODO: more tests

    # Test getting functions in subsections
    funcs = list(disassembler.functions(start=0x4003fa, end=0x40129f))
    assert len(funcs) == 3
    assert sorted(func.start for func in funcs) == [0x401000, 0x401030, 0x401150]

    # TODO: Should we be including the function where end is in the middle of the function?
    funcs = list(disassembler.functions(end=0x401049))
    assert len(funcs) == 2
    assert sorted(func.start for func in funcs) == [0x401000, 0x401030]

    # Start at almost the end so we only produce the "RtlUnwind" function.
    # (Using any earlier address will already produce inconsistent results with which functions exist)
    funcs = list(disassembler.functions(start=0x409b0c))
    assert len(funcs) == 1
    assert funcs[0].name == "RtlUnwind"
    assert funcs[0].start == 0x409b0e


def test_get_instruction(disassembler):
    addr = 0x401035

    # Test getting instruction at start.
    insn = disassembler.get_instruction(addr)
    assert insn
    assert insn.address == addr

    # Test getting instruction with address within.
    insn = disassembler.get_instruction(addr + 2)
    assert insn
    assert insn.address == addr

    # Test invalid instruction.
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_instruction(0x40c130)


def test_get_mnemonic(disassembler):
    assert disassembler.get_mnemonic(0x401035) == "push"
    assert disassembler.get_mnemonic(0x40103f) == "add"
    assert disassembler.get_mnemonic(0x4010a3) == "call"

    # Test non-instruction
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_mnemonic(0x40c130)


def test_get_operand(disassembler):
    operand = disassembler.get_operand(0x401042, 0)
    assert operand
    assert operand.address == 0x401042
    assert operand.index == 0


def test_get_operand_type(disassembler):
    operand = disassembler.get_operand(0x401042, 0)
    assert disassembler.get_operand_type(0x401042, 0) == operand.type


def test_get_operand_value(disassembler):
    operand = disassembler.get_operand(0x401042, 0)
    assert disassembler.get_operand_value(0x401042, 0) == operand.value


def test_get_byte(disassembler):
    # Code section
    assert disassembler.get_byte(0x401035) == 0x68
    assert disassembler.get_byte(0x401036) == 0x00

    # Data section
    assert disassembler.get_byte(0x40c000) == 0x49
    assert disassembler.get_byte(0x40c001) == 0x64


def test_get_bytes(disassembler):
    # Code section
    assert disassembler.get_bytes(0x401035, 2) == b"\x68\x00"

    # Data section
    assert disassembler.get_bytes(0x40c000, 2) == b"Id"


def test_find_bytes(disassembler):
    assert disassembler.find_bytes(b"\x83\xC4\x08\x33\xC0") == 0x401299
    assert disassembler.find_bytes(b"\xDE\xAD\xBE\xEF") == -1
    assert disassembler.find_bytes(b"\x83\xC4\x08\x33\xC0", start=disassembler.min_address) == 0x401299
    assert disassembler.find_bytes(b"\x83\xC4\x08\x33\xC0", start=0x4015A6) == -1
    assert disassembler.find_bytes(b"\x83\xC4\x08\x33\xC0", start=0x4015A6, reverse=True) == 0x401299
    assert disassembler.find_bytes(b"\x83\xC4\x08\x33\xC0", reverse=True) == 0x401299


# TODO: currently disabled.
# def test_get_string_bytes(disassembler):
#     # Code section
#     assert disassembler.get_string_bytes(0x401035) == b"\x68\x00"
#
#     # Data Section
#     assert disassembler.get_string_bytes(0x40c000) == b"Idmmn!Vnsme\x00"


def test_get_qword(disassembler):
    # Code section
    assert disassembler.get_qword(0x401035) == 0xffc1e80040c00068

    # Data section
    assert disassembler.get_qword(0x40c000) == 0x6e56216e6d6d6449


def test_get_dword(disassembler):
    # Code section
    assert disassembler.get_dword(0x401035) == 0x40c00068

    # Data section
    assert disassembler.get_dword(0x40c000) == 0x6d6d6449


def test_get_word(disassembler):
    # Code section
    assert disassembler.get_word(0x401035) == 0x68

    # Data section
    assert disassembler.get_word(0x40c000) == 0x6449




import pytest


@pytest.mark.parametrize("address,data", [
    # test reading loaded.
    (0x00401000, b"\x55\x8B\xEC"),
    (0x00401035, b"\x68\x00\xC0\x40\x00"),
    (0x0040C000, b"Idmmn!Vnsme \x00"),
    # TODO: Not supported in Ghidra
    # # test reading unloaded.
    # (0x0040E1AA, b"\x00\x00\x00\x00\x00"),
    # test reading partial unloaded and loaded.
    (0x0040A10c, (b"\x00" * 16) + b"\xF0\x14"),
])
def test_read(disassembler, address, data):
    with disassembler.open_memory(address, address + len(data) + 10) as memory:
        assert memory.tell_address() == address
        assert memory.read(len(data)) == data
        assert memory.tell_address() == address + len(data)
    assert disassembler.get_bytes(address, len(data)) == data


@pytest.mark.parametrize("address,data", [
    # test reading loaded.
    (0x10544, b"\x28\x10\x02\x00"),
    (0x21028, b"Idmmn!Vnsme \x00"),
])
def test_read_arm(disassembler, address, data):
    with disassembler.open_memory(address, address + len(data) + 10) as memory:
        assert memory.tell_address() == address
        assert memory.read(len(data)) == data
        assert memory.tell_address() == address + len(data)
    assert disassembler.get_bytes(address, len(data)) == data



@pytest.mark.parametrize("address", [
    # 0x00401000,  # code section  # TODO: Patching code this way doesn't work for Ghidra.
    0x0040C000,  # data section
])
def test_write(disassembler, address):
    """
    Tests patching and un-patching bytes.
    (Currently only works for data sections.)
    """
    orig_data = disassembler.get_bytes(address, 4)
    new_data = b"\xde\xad\xbe\xef"
    with disassembler.open_memory(address, address + 4) as memory:
        assert memory.write(new_data) == 4
        memory.seek(0)
        assert memory.read() == new_data

        # Ensure we can read that same data the normal way.
        assert disassembler.get_bytes(address, 4) == new_data

        # Reset
        memory.seek(0)
        assert memory.reset(4) == 4
        memory.seek(0)
        assert memory.read() == orig_data
        assert disassembler.get_bytes(address, 4) == orig_data


def test_write_uninitialized(disassembler):
    address = 0x0040D200
    with disassembler.open_memory(address, address + 4) as memory:
        with pytest.raises(IOError):
            memory.write(b"\xde\xad\xbe\xef")


def test_empty_memory(disassembler):
    with disassembler.open_memory(0x123, 0x123) as memory:
        assert memory.tell_address() == 0x123
        assert memory.read() == b""
        assert memory.tell_address() == 0x123




import pytest


# TODO: Better support pointers
@pytest.mark.parametrize("name,size", [
    ("char", 1),
    ("byte", 1),
    ("int", 4),
    ("dword", 4),
    ("word", 2),
    ("short", 2),
    ("char *", 4),
])
def test_get_data_type(disassembler, name, size):
    data_type = disassembler.get_data_type(name)
    assert data_type
    assert data_type.name == name
    assert data_type.size == size




import pytest


EXPECTED_IMPORTS_X86 = [
    (0x40a000, None, "GetCommandLineA", "KERNEL32"),
    (0x40a004, None, "EnterCriticalSection", "KERNEL32"),
    (0x40a008, None, "LeaveCriticalSection", "KERNEL32"),
    (0x40a00c, None, "TerminateProcess", "KERNEL32"),
    (0x40a010, None, "GetCurrentProcess", "KERNEL32"),
    (0x40a014, None, "UnhandledExceptionFilter", "KERNEL32"),
    (0x40a018, None, "SetUnhandledExceptionFilter", "KERNEL32"),
    (0x40a01c, None, "IsDebuggerPresent", "KERNEL32"),
    (0x40a020, None, "GetModuleHandleW", "KERNEL32"),
    (0x40a024, None, "Sleep", "KERNEL32"),
    (0x40a028, None, "GetProcAddress", "KERNEL32"),
    (0x40a02c, None, "ExitProcess", "KERNEL32"),
    (0x40a030, None, "WriteFile", "KERNEL32"),
    (0x40a034, None, "GetStdHandle", "KERNEL32"),
    (0x40a038, None, "GetModuleFileNameA", "KERNEL32"),
    (0x40a03c, None, "FreeEnvironmentStringsA", "KERNEL32"),
    (0x40a040, None, "GetEnvironmentStrings", "KERNEL32"),
    (0x40a044, None, "FreeEnvironmentStringsW", "KERNEL32"),
    (0x40a048, None, "WideCharToMultiByte", "KERNEL32"),
    (0x40a04c, None, "GetLastError", "KERNEL32"),
    (0x40a050, None, "GetEnvironmentStringsW", "KERNEL32"),
    (0x40a054, None, "SetHandleCount", "KERNEL32"),
    (0x40a058, None, "GetFileType", "KERNEL32"),
    (0x40a05c, None, "GetStartupInfoA", "KERNEL32"),
    (0x40a060, None, "DeleteCriticalSection", "KERNEL32"),
    (0x40a064, None, "TlsGetValue", "KERNEL32"),
    (0x40a068, None, "TlsAlloc", "KERNEL32"),
    (0x40a06c, None, "TlsSetValue", "KERNEL32"),
    (0x40a070, None, "TlsFree", "KERNEL32"),
    (0x40a074, None, "InterlockedIncrement", "KERNEL32"),
    (0x40a078, None, "SetLastError", "KERNEL32"),
    (0x40a07c, None, "GetCurrentThreadId", "KERNEL32"),
    (0x40a080, None, "InterlockedDecrement", "KERNEL32"),
    (0x40a084, None, "HeapCreate", "KERNEL32"),
    (0x40a088, None, "VirtualFree", "KERNEL32"),
    (0x40a08c, None, "HeapFree", "KERNEL32"),
    (0x40a090, None, "QueryPerformanceCounter", "KERNEL32"),
    (0x40a094, None, "GetTickCount", "KERNEL32"),
    (0x40a098, None, "GetCurrentProcessId", "KERNEL32"),
    (0x40a09c, None, "GetSystemTimeAsFileTime", "KERNEL32"),
    (0x40a0a0, None, "GetCPInfo", "KERNEL32"),
    (0x40a0a4, None, "GetACP", "KERNEL32"),
    (0x40a0a8, None, "GetOEMCP", "KERNEL32"),
    (0x40a0ac, None, "IsValidCodePage", "KERNEL32"),
    (0x40a0b0, 0x409b0e, "RtlUnwind", "KERNEL32"),
    (0x40a0b4, None, "LoadLibraryA", "KERNEL32"),
    (0x40a0b8, None, "InitializeCriticalSectionAndSpinCount", "KERNEL32"),
    (0x40a0bc, None, "HeapAlloc", "KERNEL32"),
    (0x40a0c0, None, "VirtualAlloc", "KERNEL32"),
    (0x40a0c4, None, "HeapReAlloc", "KERNEL32"),
    (0x40a0c8, None, "GetConsoleCP", "KERNEL32"),
    (0x40a0cc, None, "GetConsoleMode", "KERNEL32"),
    (0x40a0d0, None, "FlushFileBuffers", "KERNEL32"),
    (0x40a0d4, None, "LCMapStringA", "KERNEL32"),
    (0x40a0d8, None, "MultiByteToWideChar", "KERNEL32"),
    (0x40a0dc, None, "LCMapStringW", "KERNEL32"),
    (0x40a0e0, None, "GetStringTypeA", "KERNEL32"),
    (0x40a0e4, None, "GetStringTypeW", "KERNEL32"),
    (0x40a0e8, None, "GetLocaleInfoA", "KERNEL32"),
    (0x40a0ec, None, "SetFilePointer", "KERNEL32"),
    (0x40a0f0, None, "HeapSize", "KERNEL32"),
    (0x40a0f4, None, "CloseHandle", "KERNEL32"),
    (0x40a0f8, None, "WriteConsoleA", "KERNEL32"),
    (0x40a0fc, None, "GetConsoleOutputCP", "KERNEL32"),
    (0x40a100, None, "WriteConsoleW", "KERNEL32"),
    (0x40a104, None, "SetStdHandle", "KERNEL32"),
    (0x40a108, None, "CreateFileA", "KERNEL32"),
]


def test_imports_x86_ida(disassembler):
    actual = [
        (import_.address, import_.thunk_address, import_.name, import_.namespace)
        for import_ in disassembler.imports
    ]
    assert actual == EXPECTED_IMPORTS_X86


def test_imports_x86_ghidra(disassembler):
    # Ghidra includes the ".DLL" for the module.
    expected = [
        (address, thunk_address, name, namespace + ".DLL")
        for address, thunk_address, name, namespace in EXPECTED_IMPORTS_X86
    ]
    actual = [
        (import_.address, import_.thunk_address, import_.name, import_.namespace)
        for import_ in disassembler.imports
    ]
    assert actual == expected


def test_imports_references(disassembler):
    imp = disassembler.get_import("LoadLibraryA")
    assert set(ref.from_address for ref in imp.references_to) == {0x405873}


def test_imports_arm_ida(disassembler):
    imports = list(disassembler.imports)
    assert len(imports) == 4
    # Newer versions of IDA includes the namespace or library name.
    assert set(import_.namespace for import_ in imports) in ({".dynsym"}, {None})

    imports = [(import_.address, import_.thunk_address, import_.name) for import_ in imports]
    assert imports == [
        (0x21270, 0x102DC, "puts"),
        (0x21274, 0x10300, "abort"),
        (0x21278, 0x102E8, "__libc_start_main"),
        (0x2127C, 0x102F4, "__gmon_start__"),
    ]


def test_imports_arm_ghidra(disassembler):
    imports = [
        (import_.address, import_.thunk_address, import_.name)
        for import_ in disassembler.imports
    ]
    assert imports == [
        # Ghidra's import address is None, because it doesn't map an "extern" segment.
        (0x22004, 0x102DC, "puts"),
        (0x22008, 0x10300, "abort"),
        (0x2200c, 0x102E8, "__libc_start_main"),
        (0x22000, None, "__gmon_start__")
    ]


def test_exports_x86_ida(disassembler):
    exports = [(export.address, export.name) for export in disassembler.exports]
    assert exports == [(0x4014E0, "start")]


def test_exports_x86_ghidra(disassembler):
    exports = [(export.address, export.name) for export in disassembler.exports]
    assert exports == [(0x4014E0, "entry")]


def test_exports_arm(disassembler):
    # NOTE: Only testing main user code symbols, since the other entry points
    # can vary based on disassembler.
    expected = [
        # (0x102bc, ".init_proc"),
        # (0x1030c, "_start"),
        (0x103fc, "encrypt"),
        (0x1045c, "decrypt"),
        (0x1058c, "main"),
        # (0x1067c, "__libc_csu_init"),
        # (0x106dc, "__libc_csu_fini"),
        # (0x106e0, ".term_proc"),
        # (0x106e8, "_IO_stdin_used"),
        # (0x21020, "__data_start"),
        # (0x21024, "__dso_handle"),
        (0x21028, "string01"),
        (0x21038, "string02"),
        (0x21054, "string03"),
        (0x21084, "string04"),
        (0x210a8, "string05"),
        (0x210c8, "string06"),
        (0x210ec, "string13"),
        (0x21118, "string17"),
        (0x2113c, "string1a"),
        (0x21148, "string23"),
        (0x21158, "string27"),
        (0x21160, "string40"),
        (0x21168, "string46"),
        (0x21184, "string73"),
        (0x2119c, "string75"),
        (0x211c4, "string77"),
        (0x211ec, "string7a"),
        (0x2121c, "string7f"),
        # (0x2126a, "__bss_start"),

    ]
    actual = sorted([
        (export.address, export.name)
        for export in disassembler.exports
        if export.name in ("encrypt", "decrypt", "main") or export.name.startswith("string")
    ])
    assert actual == expected




import pytest

import dragodis


# TODO: Find and test more complex examples, such as fragmented functions or functions that
#   don't start at minimum address.
# TODO: Test a function where the last instruction is more than 1 bytes to ensure proper
#   end address is realized.
@pytest.mark.parametrize("address,start,end", [
    (0x401030, 0x401030, 0x401143),
    (0x401030 + 10, 0x401030, 0x401143),
    (0x401beb, 0x40189c, 0x402413),
])
def test_start_end(disassembler, address, start, end):
    func = disassembler.get_function(address)
    assert func
    assert func.start == start
    assert func.end == end
    assert address in func


def test_name(disassembler):
    # Test a library function, which should have the same name regardless of disassembler.
    func = disassembler.get_function(0x4012a0)
    assert func.name == "_printf"
    func.name = "new_func_name"
    assert func.name == "new_func_name"
    func.name = None
    assert func.name != "new_func_name"
    # disassembler should reset default naming scheme since we lose the analysis information.
    assert "4012" in func.name


def test_comment(disassembler):
    func = disassembler.get_function(0x401030)
    func.set_comment("regular comment")
    assert func.get_comment() == "regular comment"
    func.set_comment("repeatable comment", dragodis.CommentType.repeatable)
    assert func.get_comment() == "regular comment"
    assert func.get_comment(dragodis.CommentType.repeatable) == "repeatable comment"
    func.set_comment(None)
    assert func.get_comment() is None
    func.set_comment(None, dragodis.CommentType.repeatable)
    assert func.get_comment(dragodis.CommentType.repeatable) is None

    # Test invalid comment type for a function.
    with pytest.raises(ValueError):
        func.set_comment("test", dragodis.CommentType.posterior)


def test_references(disassembler):
    func = disassembler.get_function(0x401000)
    refs = list(func.references_to)
    # Number of references can be 18 or 20 depending if the disassembler counts the referent in the PE header due to being the first function in the text section.
    assert len(refs) in (18, 20)
    assert all(ref.to_address == 0x401000 for ref in refs)
    assert sorted(ref.from_address for ref in refs if ref.is_code) == [
        0x40103a,
        0x401049,
        0x401058,
        0x401067,
        0x401076,
        0x401085,
        0x401094,
        0x4010a3,
        0x4010b2,
        0x4010c1,
        0x4010d0,
        0x4010df,
        0x4010ee,
        0x4010fd,
        0x40110c,
        0x40111b,
        0x40112a,
        0x401139,
    ]

    refs = list(func.references_from)
    assert len(refs) == 0


def test_call_references(disassembler):
    func = disassembler.get_function(0x401000)
    assert list(func.calls_to) == [
        0x40103a,
        0x401049,
        0x401058,
        0x401067,
        0x401076,
        0x401085,
        0x401094,
        0x4010a3,
        0x4010b2,
        0x4010c1,
        0x4010d0,
        0x4010df,
        0x4010ee,
        0x4010fd,
        0x40110c,
        0x40111b,
        0x40112a,
        0x401139,
    ]
    assert list(func.calls_from) == []
    callers = list(func.callers)
    assert len(callers) == 1
    assert callers[0].start == 0x401030
    assert list(func.callees) == []

    func = disassembler.get_function(0x401030)
    assert list(func.calls_to) == [
        0x401153,
    ]
    assert list(func.calls_from) == [
        (0x40103a, 0x401000),
        (0x401049, 0x401000),
        (0x401058, 0x401000),
        (0x401067, 0x401000),
        (0x401076, 0x401000),
        (0x401085, 0x401000),
        (0x401094, 0x401000),
        (0x4010a3, 0x401000),
        (0x4010b2, 0x401000),
        (0x4010c1, 0x401000),
        (0x4010d0, 0x401000),
        (0x4010df, 0x401000),
        (0x4010ee, 0x401000),
        (0x4010fd, 0x401000),
        (0x40110c, 0x401000),
        (0x40111b, 0x401000),
        (0x40112a, 0x401000),
        (0x401139, 0x401000),
    ]
    callers = list(func.callers)
    assert len(callers) == 1
    assert callers[0].start == 0x401150
    callees = list(func.callees)
    assert len(callees) == 1
    assert callees[0].start == 0x401000


def test_instructions(disassembler):
    func = disassembler.get_function(0x401000)
    instructions = list(func.instructions())
    assert [instruction.address for instruction in instructions] == [
        0x401000, 0x401001, 0x401003, 0x401006, 0x401009, 0x40100b,
        0x40100d, 0x401011, 0x401014, 0x401017, 0x401019, 0x40101c,
        0x40101e, 0x401021, 0x401024, 0x401027, 0x401029, 0x40102a,
    ]
    instructions = list(func.instructions(start=0x40101e))
    assert [instruction.address for instruction in instructions] == [
        0x40101e, 0x401021, 0x401024, 0x401027, 0x401029, 0x40102a,
    ]
    instructions = list(func.instructions(end=0x40100b))
    assert [instruction.address for instruction in instructions] == [
        0x401000, 0x401001, 0x401003, 0x401006, 0x401009,
    ]
    instructions = list(func.instructions(start=0x40100b, end=0x40101e))
    assert [instruction.address for instruction in instructions] == [
        0x40100b, 0x40100d, 0x401011, 0x401014, 0x401017, 0x401019,
        0x40101c,
    ]


def test_source_code_ghidra(disassembler):
    func = disassembler.get_function(0x401000)
    assert func.source_code == """
void __cdecl FUN_00401000(byte *param_1,byte param_2)

{
  for (; *param_1 != 0; param_1 = param_1 + 1) {
    *param_1 = *param_1 ^ param_2;
  }
  return;
}

"""


def test_source_code_ida(disassembler):
    func = disassembler.get_function(0x401000)
    assert func.source_code == """\
_BYTE *__cdecl sub_401000(_BYTE *a1, char a2)
{
  _BYTE *result; // eax

  while ( 1 )
  {
    result = a1;
    if ( !*a1 )
      break;
    *a1++ ^= a2;
  }
  return result;
}
"""


def test_data(disassembler):
    func = disassembler.get_function(0x401000)
    assert func.data == (
        b"\x55"
        b"\x8B\xEC"
        b"\x8B\x45\x08"
        b"\x0F\xBE\x08"
        b"\x85\xC9"
        b"\x74\x1C"
        b"\x0F\xBE\x55\x0C"
        b"\x8B\x45\x08"
        b"\x0F\xBE\x08"
        b"\x33\xCA"
        b"\x8B\x55\x08"
        b"\x88\x0A"
        b"\x8B\x45\x08"
        b"\x83\xC0\x01"
        b"\x89\x45\x08"
        b"\xEB\xDA" 
        b"\x5D"
        b"\xC3"
    )


@pytest.mark.parametrize("address,result", [
    (0x401000, False),
    (0x4012a0, True),
])
def test_is_library(disassembler, address, result):
    func = disassembler.get_function(address)
    assert func.is_library == result




import pytest

from dragodis import SegmentType, SegmentPermission


def test_basic(disassembler):
    if disassembler.name.lower() == "ida":
        expected_segments = [
            (0x00401000, 0x0040A000, ".text", True, SegmentPermission.read | SegmentPermission.execute),
            (0x0040A000, 0x0040A110, ".idata", False, SegmentPermission.read),
            (0x0040A110, 0x0040C000, ".rdata", True, SegmentPermission.read),
            (0x0040C000, 0x0040F000, ".data", True, SegmentPermission.read | SegmentPermission.write),
        ]
    elif disassembler.name.lower() == "ghidra":
        expected_segments = [
            (0x00400000, 0x00400400, "Headers", True, SegmentPermission.read),
            (0x00401000, 0x00409C00, ".text", True, SegmentPermission.read | SegmentPermission.execute),
            (0x0040A000, 0x0040BE00, ".rdata", True, SegmentPermission.read),
            (0x0040C000, 0x0040D200, ".data", True, SegmentPermission.read | SegmentPermission.write),
            (0x0040D200, 0x0040ed48, ".data", False, SegmentPermission.read | SegmentPermission.write),
        ]
    else:
        raise NotImplementedError

    actual_segments = list(disassembler.segments)
    # Ghidra 10.2.* sometimes includes an extra "tdb" memory block... we are just going ignore that.
    actual_segments = [seg for seg in actual_segments if seg.name != "tdb"]
    print(actual_segments)
    assert len(actual_segments) == len(expected_segments)
    for segment, expected_segments in zip(actual_segments, expected_segments):
        start, end, name, initialized, permissions = expected_segments
        assert segment
        assert hex(segment.start) == hex(start)
        assert hex(segment.end) == hex(end)
        assert segment.name == name
        assert segment.initialized == initialized
        assert segment.permissions == permissions
        assert start in segment
        assert start + 1 in segment

        # Test direct retrieval
        # Doesn't work for the Ghidra segment duplicate.
        if name == ".data" and not initialized:
            continue
        segment2 = disassembler.get_segment(name)
        assert segment2 == segment
        segment3 = disassembler.get_segment(start)
        assert segment3 == segment


def test_create_segment(disassembler):
    orig_segments = list(disassembler.segments)
    segment = disassembler.create_segment(".test", 0x1234, 256)
    assert segment
    assert segment.name == ".test"
    assert segment.start == 0x1234
    assert segment.end == 0x1234 + 256
    assert not segment.initialized

    segments = list(disassembler.segments)
    assert len(segments) == len(orig_segments) + 1
    assert any(seg.name == ".test" for seg in segments)


def test_data(disassembler):
    # Test pulling from loaded memory.
    segment = disassembler.get_segment(".text")
    with segment.open() as memory:
        assert memory.read(0x20) == (
            b"\x55\x8B\xEC\x8B\x45\x08\x0F\xBE\x08\x85\xC9\x74\x1C\x0F\xBE\x55"
            b"\x0C\x8B\x45\x08\x0F\xBE\x08\x33\xCA\x8B\x55\x08\x88\x0A\x8B\x45"
        )

        # Test seeking and then pulling different data.
        memory.seek_address(0x00401141)
        assert memory.read(4) == b"\x5d\xc3\xcc\xcc"

    assert segment.get_bytes(0x00401141, 4) == b"\x5d\xc3\xcc\xcc"




import pytest

from dragodis.interface import Register


@pytest.mark.parametrize("address,index,name,width", [
    (0x401001, 0, "ebp", 32),      # mov     ebp, esp
    (0x401009, 0, "ecx", 32),      # test    ecx, ecx
    (0x40101C, 1, "cl", 8),        # mov     [edx], cl
    (0x402F73, 0, "esi", 32),      # add     esi, ebx
])
def test_basic(disassembler, address, index, name, width):
    register = disassembler.get_operand_value(address, index)
    assert isinstance(register, Register)
    assert register.name == name
    assert register.bit_width == width


def test_get_register(disassembler):
    reg = disassembler.get_register("ecx")
    assert isinstance(reg, Register)
    assert reg.name == "ecx"
    assert reg.bit_width == 32


def test_equality(disassembler):
    # test    ecx, ecx
    ecx1 = disassembler.get_operand_value(0x401009, 0)
    ecx2 = disassembler.get_operand_value(0x401009, 1)
    assert ecx1 == ecx2

    # add     esi, ebx
    esi = disassembler.get_operand_value(0x402F73, 0)
    ebx = disassembler.get_operand_value(0x402F73, 1)
    assert esi != ebx

    # 0x401009: test    ecx, ecx
    # 0x40101C: mov     [edx], cl
    ecx = disassembler.get_operand_value(0x401009, 0)
    cl = disassembler.get_operand_value(0x40101C, 1)
    assert ecx != cl


@pytest.mark.parametrize("register,mask", [
    ("al", 0xFF),
    ("ah", 0xFF00),
    ("eax", 0xFFFFFFFF),
    ("eip", 0xFFFFFFFF),
    ("esp", 0xFFFFFFFF)
])
def test_mask(disassembler, register, mask):
    reg = disassembler.get_register(register)
    assert reg.mask == mask


@pytest.mark.parametrize("register,base", [
    ("al", "eax"),
    ("ah", "eax"),
    ("eax", "eax"),
    ("eip", "eip"),
    ("esp", "esp"),
    ("fs", "fs")
])
def test_base(disassembler, register, base):
    reg = disassembler.get_register(register)
    assert reg.base == disassembler.get_register(base)




import pytest


def _test_stack_frame(disassembler, address, variables):
    function = disassembler.get_function(address)
    stack_frame = function.stack_frame
    assert len(stack_frame) == len(variables)
    # First test iterating through members.
    actual_vars = [(var.stack_offset, var.name, var.data_type.name) for var in stack_frame]
    assert actual_vars == variables
    # Then test obtaining them through indexing.
    for stack_offset, name, data_type in variables:
        var = stack_frame[stack_offset]
        assert var.stack_offset == stack_offset
        assert var.name == name
        var = stack_frame[name]
        assert var.stack_offset == stack_offset
        assert var.name == name


@pytest.mark.parametrize("address,variables", [
    (0x401000, [
        (4, "arg_0", "int"),
        (8, "arg_4", "char"),
    ]),
    (0x4044F4, [
        (-0x14, "var_10", "int"),
        (-0xc, "var_8", "int"),
        (-8, "var_4", "char"),
    ])
])
def test_stack_frame_ida(disassembler, address, variables):
    _test_stack_frame(disassembler, address, variables)


@pytest.mark.parametrize("address,variables", [
    (0x103FC, [
        (-0xd, "var_9", "char"),
        (-0xc, "var_8", "int"),
        # IDA is missing variable at -4 since it interprets PUSH mnemonics instead of
        # explicitly defining the STR instruction.
        # e.g:  str  r11,[sp,#local_4]!
    ])
])
def test_stack_frame_arm_ida(disassembler, address, variables):
    _test_stack_frame(disassembler, address, variables)


@pytest.mark.parametrize("address,variables", [
    (0x401000, [
        (4, "param_1", "byte *"),
        (8, "param_2", "byte"),
    ]),
    (0x4044F4, [
        (-0x14, "local_14", "undefined4"),
        (-0xc, "local_c", "undefined4"),
        (-8, "local_8", "undefined1"),
        (4, "param_1", "int"),  # (Ghidra also thinks there is an argument)
    ])
])
def test_stack_frame_ghidra(disassembler, address, variables):
    _test_stack_frame(disassembler, address, variables)


@pytest.mark.parametrize("address,variables", [
    (0x103FC, [
        (-0xd, "local_d", "undefined1"),
        (-0xc, "local_c", "undefined4"),
        (-4, "local_4", "undefined4"),
    ])
])
def test_stack_frame_arm_ghidra(disassembler, address, variables):
    _test_stack_frame(disassembler, address, variables)




import pytest

import dragodis
from dragodis import OperandType
from dragodis.interface import Register, Phrase, StackVariable, GlobalVariable, RegisterList
from dragodis.interface.types import ARMShiftType


def test_basic(disassembler):
    """
    Tests basic operand properties and general smoke test.
    """
    operand = disassembler.get_operand(0x401042, 0)
    assert operand
    assert operand.address == 0x401042
    assert operand.index == 0
    assert operand.type == OperandType.immediate
    assert operand.value == 2

    operand = disassembler.get_operand(0x40103f, 1)
    assert operand.address == 0x40103f
    assert operand.index == 1
    assert operand.type == OperandType.immediate
    assert operand.value == 8

    # Test invalid index
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_operand(0x401042, 3)

    # Test non-instruction
    with pytest.raises(dragodis.NotExistError):
        disassembler.get_operand(0x40c130, 0)


def _test_operand(disassembler, address, index, text, width):
    """
    Tests core functionality of operand.
    """
    operand = disassembler.get_operand(address, index)
    assert operand.address == address
    assert operand.index == index
    assert operand.text == text
    assert str(operand) == text
    assert operand.width == width


@pytest.mark.parametrize("address,index,text,width", [
    (0x401042, 0, "2", 4),
    (0x40135f, 0, "cl", 1),
    (0x40103f, 0, "esp", 4),
    (0x401044, 0, "offset aVgqvQvpkleUkvj", 4),
    (0x40177a, 0, "byte ptr [esi+0Ch]", 1),
    (0x40177a, 1, "0", 1),  # 1 because the operand it goes into is 1
    (0x4012fe, 1, "[ebp+arg_4]", 4),
    (0x401006, 1, "byte ptr [eax]", 1),
    (0x401003, 1, "[ebp+arg_0]", 4),
])
def test_text_ida(disassembler, address, index, text, width):
    _test_operand(disassembler, address, index, text, width)


@pytest.mark.parametrize("address,index,text,width", [
    (0x401042, 0, "0x2", 4),
    (0x40135f, 0, "CL", 1),
    (0x40103f, 0, "ESP", 4),
    (0x401044, 0, "0x40c010", 4),
    (0x40177a, 0, "byte ptr [ESI + 0xc]", 1),
    (0x40177a, 1, "0x0", 1),  # 1 because the operand it goes into is 1
    (0x4012fe, 1, "[EBP + 0xc]", 4),
    (0x401006, 1, "byte ptr [EAX]", 1),
    (0x401003, 1, "dword ptr [EBP + 0x8]", 4),
])
def test_text_ghidra(disassembler, address, index, text, width):
    _test_operand(disassembler, address, index, text, width)


# TODO: Add more tests as we solve questions brought up in issue #12
def test_value_x86(disassembler):
    # push    2
    # PUSH    0x2
    immediate = disassembler.get_operand_value(0x401042, 0)
    assert immediate == 2
    assert int(immediate) == 2

    # add     esp, 8
    # ADD     ESP,0x8
    immediate = disassembler.get_operand_value(0x40103f, 1)
    assert immediate == 8
    assert int(immediate) == 8
    register = disassembler.get_operand_value(0x40103f, 0)
    assert register and isinstance(register, Register)
    assert register.name == "esp"
    assert register.bit_width == 32
    assert int(register) == -1

    # push    offset aVgqvQvpkleUkvj
    # PUSH    s_Vgqv"qvpkle"ukvj"ig{"2z20_0040c010
    address = disassembler.get_operand_value(0x401044, 0)
    assert address == 0x40c010
    assert int(address) == 0x40c010

    # lea     eax, [ebp+arg_4]
    # LEA     EAX=>Stack[0x8],[EBP + 0xc]
    phrase = disassembler.get_operand_value(0x4012fe, 1)
    assert phrase and isinstance(phrase, Phrase)
    assert phrase.offset == 0xc
    assert phrase.base.name == "ebp"
    assert int(phrase) == 0xc

    # call    sub_401000
    address = disassembler.get_operand_value(0x4010a3, 0)
    assert address == 0x401000
    assert int(address) == 0x401000

    # add     ecx, dword_40DC20[esi*4]
    # ADD     ECX,dword ptr [ESI*0x4 + DAT_0040dc20]
    phrase = disassembler.get_operand_value(0x401965, 1)
    assert phrase and isinstance(phrase, Phrase)
    assert phrase.base is None
    assert phrase.index.name == "esi"
    assert phrase.scale == 4
    assert phrase.offset == 0x40dc20
    assert int(phrase) == 0x40dc20


def test_value_arm(disassembler):
    # Second operand value should be a register type with a shift.
    # movs       r6,r6, asr #0x2
    operand = disassembler.get_operand(0x106A4, 1)
    assert operand.type == OperandType.register
    assert operand.shift == (ARMShiftType.ASR, 2)
    value = operand.value
    assert value and isinstance(value, Register)
    assert value.name == "r6"
    assert operand.width == 4
    assert int(value) == -1


def _test_phrase(disassembler, address, index, phrase):
    operand = disassembler.get_operand(address, index)
    assert operand.type == OperandType.phrase
    value = operand.value
    assert isinstance(value, Phrase)
    base, index, scale, offset = phrase

    if not base:
        assert value.base is None
    else:
        assert isinstance(value.base, Register)
        assert value.base.name == base

    if not index:
        assert value.index is None
    else:
        assert isinstance(value.index, Register)
        assert value.index.name == index

    assert value.scale == scale

    if isinstance(offset, str):
        assert isinstance(value.offset, Register)
        assert value.offset.name == offset
    else:
        assert value.offset == offset


@pytest.mark.parametrize("address,index,phrase", [
    (0x4012fe, 1, ("ebp", None, 1, 0xc)),  # [EBP + 0xc]
    (0x401313, 0, ("ebp", None, 1, -0x1c)),  # dword ptr [EBP + -0x1c]=>local_2
    # TODO: Not a valid phrase, its a register with width of 1
    # (0x401006, 0, ("eax", None, 1, 0)),  # byte ptr [EAX]
    (0x4013be, 0, ("eax", None, 1, 0x400018)),  # word ptr [EAX + 0x400018]
    (0x40154e, 0, ("edx", "eax", 1, 0)),  # dword ptr [EDX + EAX*0x1]
    # TODO: Figure out what GHidra is typing this and then get it to a phrase.
    (0x40156f, 1, (None, "eax", 4, 0x40dc20)),  # dword ptr [EAX*0x4 + DAT_0040dc20]
])
def test_phrase_x86(disassembler, address, index, phrase):
    _test_phrase(disassembler, address, index, phrase)


@pytest.mark.parametrize("address,index,phrase", [
    (0x10418, 1, ("r11", None, 1, -8)),     # [r11,#-0x8]=>local_c
    (0x102f0, 1, ("r12", None, 1, 0xd20)),  # [r12,#0xd20]!=>->__libc_start_main
    (0x10354, 1, ("r3", None, 1, "r2")),    # [r3,r2]=>->__gmon_start__
])
def test_phrase_arm(disassembler, address, index, phrase):
    _test_phrase(disassembler, address, index, phrase)


@pytest.mark.parametrize("address,index,names", [
    # PUSH    {R11,LR}
    # stmdb      sp!,{r11 lr}
    (0x1058C, 0, ["sp", "r11", "lr"]),
    # POP     {R4-R10,PC}
    # ldmia      sp!,{r4 r5 r6 r7 r8 r9 r10 pc}
    (0x106D0, 0, ["sp", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "pc"]),
])
def test_register_list_arm(disassembler, address, index, names):
    """
    Special tests for register list operands.
    """
    instruction = disassembler.get_instruction(address)

    # IDA interprets stmdb and ldmia instructions as "PUSH" and "POP"
    # with the first "sp" register implied.
    if instruction.mnemonic in ("push", "pop") and names[0] == "sp":
        names = names[1:]

    operand = instruction.operands[0]
    assert operand.type == OperandType.register_list
    reg_list = operand.value
    assert isinstance(reg_list, RegisterList)
    assert [reg.name for reg in reg_list] == names
    assert int(reg_list) == -1


@pytest.mark.parametrize("address,index,operand_type", [
    (0x401042, 0, OperandType.immediate),   # 2
    (0x40103f, 0, OperandType.register),    # esp
    (0x40103f, 1, OperandType.immediate),   # 8
    # Offsets are treated as immediate unless explicitly dereferenced by the instruction.
    (0x401044, 0, OperandType.immediate),   # offset aVgqvQvpkleUkv
    (0x4015D3, 1, OperandType.immediate),   # offset unk_40C4F8
    (0x40243C, 0, OperandType.memory),      # dword_40D19C  /  [DAT_0040d19c]
    (0x4012fe, 1, OperandType.phrase),      # [ebp+arg_4]
    (0x40156f, 1, OperandType.phrase),      # dword ptr [EAX*0x4 + DAT_0040dc20]
])
def test_type_x86(disassembler, address, index, operand_type):
    operand = disassembler.get_operand(address, index)
    assert operand.type == operand_type


@pytest.mark.parametrize("address,index,operand_type", [
    (0x106A4, 1, OperandType.register),       # r6, asr #0x2
    (0x1067C, 0, OperandType.register_list),  # sp!,{r4 r5 r6 r7 r8 r9 r10 lr}
    (0x10464, 1, OperandType.immediate),      # #1
    (0x10468, 1, OperandType.memory),         # =string01  /  [->string01]
    (0x10334, 0, OperandType.code),           # __libc_start_main
])
def test_type_arm(disassembler, address, index, operand_type):
    operand = disassembler.get_operand(address, index)
    assert operand.type == operand_type


@pytest.mark.parametrize("address,index,operand_type", [
    (0x103fc, 0, OperandType.register_list),  # {R11}
])
def test_type_ida_arm(disassembler, address, index, operand_type):
    operand = disassembler.get_operand(address, index)
    assert operand.type == operand_type


# TODO: Why is the operand being seen as a register_list?

@pytest.mark.parametrize("address,index,operand_type", [
    (0x103fc, 0, OperandType.register),  # r11
    (0x103fc, 1, OperandType.phrase),    # [sp,#local_4]!
])
def test_type_ghidra_arm(disassembler, address, index, operand_type):
    operand = disassembler.get_operand(address, index)
    assert operand.type == operand_type


@pytest.mark.parametrize("address,index,shift_type,shift_count", [
    (0x106A4, 1, ARMShiftType.ASR, 2),   # R6,ASR#2
    (0x103A4, 1, ARMShiftType.LSR, 31),  # R3,LSR#31
    (0x10698, 1, ARMShiftType.LSL, 0),   # R1  (no shift, default)
])
def test_shift_info_arm(disassembler, address, index, shift_type, shift_count):
    instruction = disassembler.get_instruction(address)
    operand = instruction.operands[index]
    assert operand.shift == (shift_type, shift_count)


def _test_variable(disassembler, address, index, name, size, data_type, location):
    instruction = disassembler.get_instruction(address)
    operand = instruction.operands[index]
    variable = operand.variable
    assert variable
    assert variable.name == name
    assert variable.size == size

    data_type_name = variable.data_type.name
    # Ghidra 10.4 switch 'undefined' to 'undefined1'
    if data_type_name == "undefined1":
        data_type_name = "undefined"
    assert data_type_name == data_type

    if isinstance(variable, GlobalVariable):
        assert variable.address == location
    elif isinstance(variable, StackVariable):
        assert variable.stack_offset == location
    else:
        pytest.fail(f"Unexpected variable type: {type(variable)}")


@pytest.mark.parametrize("address,index,name,size,data_type,location", [
    (0x40100d, 1, "arg_4", 1, "char", 8),
    (0x401024, 0, "arg_0", 4, "int", 4),
    (0x4058b4, 0, "lpProcName", 4, "lpcstr", -0x28),
    (0x4058b4, 1, "aGetlastactivep", 19, "char", 0x40a9a0),
    (0x4015a6, 0, "byte_40D1C8", 1, "byte", 0x40d1c8),
    (0x404096, 0, "LeaveCriticalSection", 4, "dword", 0x40A008),
    # TODO: Determine if we should support pointers to code as "variables".
    # (0x40103A, 0, "sub_401000", 4, "dword", 0x401000),
    # (0x40100b, 0, "loc_401029", 1, "byte", 0x401029),
])
def test_variable_ida(disassembler, address, index, name, size, data_type, location):
    _test_variable(disassembler, address, index, name, size, data_type, location)


@pytest.mark.parametrize("address,index,name,size,data_type,location", [
    (0x40100d, 1, "param_2", 1, "byte", 8),
    (0x401024, 0, "param_1", 4, "byte *", 4),
    (0x40585b, 0, "local_8", 4, "undefined4", -0x8),
    pytest.param(
        0x4058b4, 0, "local_28", 4, "undefined4", -0x28,
        marks=pytest.mark.xfail(reason="Ghidra 10.3.2 sometimes doesn't make this variable.")
    ),
    (0x4058b4, 1, "s_GetLastActivePopup_0040a9a0", 19, "string", 0x40a9a0),
    (0x4015a6, 0, "DAT_0040d1c8", 1, "undefined", 0x40d1c8),
    (0x404096, 0, "PTR_LeaveCriticalSection_0040a008", 4, "pointer", 0x40A008),
    # (0x40103A, 0, "FUN_00401000", 4, "dword", 0x401000),
    # (0x40100b, 0, "LAB_00401029", 1, "byte", 0x401029),
])
def test_variable_ghidra(disassembler, address, index, name, size, data_type, location):
    _test_variable(disassembler, address, index, name, size, data_type, location)




def test_basic(disassembler):
    strings = list(disassembler.strings())
    assert set((entry.address, str(entry)) for entry in strings) >= {
        # A handful of the expected strings we should see.
        (0x40b208, "MM/dd/yy"),
        (0x40a96c, "GetProcessWindowStation"),
        (0x40ca56, "ABCDEFGHIJKLMNOPQRSTUVWXYZ"),
        (0x40c000, "Idmmn!Vnsme "),
    }

    for string in strings:
        if string.address == 0x40c000:
            assert bytes(string) == b"Idmmn!Vnsme "




import pytest

import dragodis


# TODO: Generalize the testing of .text since the actual string could change lot based
#    on options set in the disassembler.
from dragodis import FlowType
from dragodis.interface.instruction import ARMConditionCode


@pytest.mark.parametrize("address,mnemonic,text", [
    (0x401000, "push", {
        "ida": "push    ebp",
        "ghidra": "PUSH EBP",
    }),
    (0x401003, "mov", {
        "ida": "mov     eax, [ebp+arg_0]",
        "ghidra": "MOV EAX,dword ptr [EBP + 0x8]",
    })
])
def test_basic(disassembler, address, mnemonic, text):
    text = text[disassembler.name.lower()]
    instruction = disassembler.get_instruction(address)

    assert instruction.address == address
    assert instruction.mnemonic == mnemonic
    assert instruction.text == text


@pytest.mark.parametrize("address,num_operands", [
    (0x401030, 1),
    (0x401031, 2),
    (0x401142, 0),
    (0x40655e, 1),  # Hidden operand?
    (0x406623, 1),
])
def test_operands(disassembler, address, num_operands):
    instruction = disassembler.get_instruction(address)
    operands = instruction.operands
    assert len(operands) == num_operands
    for index, operand in enumerate(operands):
        assert operand.address == address
        assert operand.index == index


# TODO: Instead having these is_*() type functions, we should probably have
#   an "InstructionType" or "FlowType" that tells the user what type of flow the function
#    has?
@pytest.mark.parametrize("address,flow_type", [
    (0x401003, FlowType.fall_through),
    (0x40103a, FlowType.call),
    (0x40100b, FlowType.conditional_jump),
    (0x401027, FlowType.unconditional_jump),
    (0x40102a, FlowType.terminal),
])
def test_flow_type(disassembler, address, flow_type):
    instruction = disassembler.get_instruction(address)
    assert instruction.flow_type == flow_type


@pytest.mark.parametrize("address,root_mnem", [
    (0x40100D, "movsx"),
    # original has "rep movsd" or "MOVSD.REP"
    # NOTE: not requiring the "d" to be removed since that is not possible for Ghidra to detect.
    (0x405c5a, ("movsd", "movs")),
])
def test_root_mnemonic_x86(disassembler, address, root_mnem):
    instruction = disassembler.get_instruction(address)
    if isinstance(root_mnem, tuple):
        assert instruction.root_mnemonic in root_mnem
    else:
        assert instruction.root_mnemonic == root_mnem


@pytest.mark.parametrize("address,root_mnem", [
    (0x10444, "b"),  # from "bne"
    (0x10424, "ldr"),  # from "ldrb"
])
def test_root_mnemonic_arm(disassembler, address, root_mnem):
    instruction = disassembler.get_instruction(address)
    assert instruction.root_mnemonic == root_mnem


@pytest.mark.parametrize("address,rep", [
    (0x4047aa, "rep"),  # rep movsd
    (0x408590, "rep"),  # rep stosd
    (0x40858a, None),   # mov
])
def test_rep_x86(disassembler, address, rep):
    instruction = disassembler.get_instruction(address)
    assert instruction.rep == rep


@pytest.mark.parametrize("address,condition_code", [
    (0x106cc, ARMConditionCode.NE),  # bne
    (0x103e0, ARMConditionCode.NE),  # ldmiane  (popne for IDA)
    (0x103b0, ARMConditionCode.EQ),  # bxeq
    (0x1039c, ARMConditionCode.AL),  # ldr
])
def test_condition_codes_arm(disassembler, address, condition_code):
    instruction = disassembler.get_instruction(address)
    assert instruction.condition_code == condition_code


@pytest.mark.parametrize("address,writeback,pre_indexed,post_indexed", [
    (0x1069C, False, False, False),  # MOV     R9, R2
    (0x106bc, True, False, True),    # LDR     R3, [R5],#4
    (0x102D4, True, True, False),    # LDR     PC, [LR,#8]!
    # TODO: Figure this one out.
    (0x103fc, True, True, False),    # PUSH    {R11}         / str     r11,[sp,#local_4]!
    (0x102C8, True, True, False),    # PUSH    {LR}          / str        lr,[sp,#-0x4]!
    (0x106d0, True, True, False),    # POP     {R4-R10,PC}   / ldmia      sp!,{r4 r5 r6 r7 r8 r9 r10 pc}])
])
def test_writeback_arm(disassembler, address, writeback, pre_indexed, post_indexed):
    instruction = disassembler.get_instruction(address)
    assert instruction.writeback == writeback
    assert instruction.pre_indexed == pre_indexed
    assert instruction.post_indexed == post_indexed


@pytest.mark.parametrize("address,stack_depth,stack_delta", [
    (0x401030, 0, -4),  # beginning of function
    (0x401031, -4, 0),
    (0x401033, -4, -4),
    (0x401035, -8, -4),
    (0x40103a, -0xc, 0),
    (0x40103f, -0xc, 8),
    (0x401042, -4, -4),
    (0x401142, 0, 0),   # end of function
])
def test_stack_depth_delta(disassembler, address, stack_depth, stack_delta):
    instruction = disassembler.get_instruction(address)
    assert instruction.stack_depth == stack_depth
    assert instruction.stack_delta == stack_delta




import pytest

from dragodis import NotExistError
from dragodis.interface.function_argument_location import StackLocation


def test_signature(disassembler):
    func = disassembler.get_function(0x401000)
    signature = func.signature
    assert signature.name
    assert signature.name == func.name


def test_signature_for_import(disassembler):
    """
    Tests getting a function signature for an external import.
    """
    address = 0x40a0c4
    # First show, we can't get the function normally.
    with pytest.raises(NotExistError):
        disassembler.get_function(address)
    # Now show we can still get the function signature because it is an import.
    signature = disassembler.get_function_signature(address)
    assert signature.name == "HeapReAlloc"


@pytest.mark.parametrize("address,declaration", [
    (0x401000, "_BYTE *__cdecl sub_401000(_BYTE *a1, char a2);"),
    (0x40a0c4, "LPVOID __stdcall HeapReAlloc(HANDLE hHeap, DWORD dwFlags, LPVOID lpMem, SIZE_T dwBytes);")
])
def test_signature_declaration_ida(disassembler, address, declaration):
    signature = disassembler.get_function_signature(address)
    assert signature.declaration == declaration


@pytest.mark.parametrize("address,declaration", [
    (0x401000, "undefined __cdecl FUN_00401000(byte * param_1, byte param_2)"),
    (0x40a0c4, "LPVOID __stdcall HeapReAlloc(HANDLE hHeap, DWORD dwFlags, LPVOID lpMem, SIZE_T dwBytes)")
])
def test_signature_declaration_ghidra(disassembler, address, declaration):
    signature = disassembler.get_function_signature(address)
    assert signature.declaration == declaration


@pytest.mark.parametrize("address,calling_convention", [
    (0x401000, "__cdecl"),
    (0x40a0c4, "__stdcall"),
])
def test_calling_convention(disassembler, address, calling_convention):
    signature = disassembler.get_function_signature(address)
    # test getting calling convention
    assert signature.calling_convention == calling_convention
    # Ghidra doesn't include the "__" in declaration
    assert calling_convention.lstrip("_") in signature.declaration
    # test changing calling convention
    signature.calling_convention = "fastcall"
    assert signature.calling_convention == "__fastcall"
    assert "fastcall" in signature.declaration
    # reset
    signature.calling_convention = calling_convention
    assert signature.calling_convention == calling_convention


@pytest.mark.parametrize("address,return_types", [
    (0x401150, ("int", "undefined4")),
    (0x40a0c4, ("lpvoid", "LPVOID")),
])
def test_return_type(disassembler, address, return_types):
    signature = disassembler.get_function_signature(address)
    # test getting the return type
    orig_type = signature.return_type
    assert str(orig_type) in return_types
    # test setting the return type
    signature.return_type = "char *"
    assert str(signature.return_type) == "char *"
    # reset and test setting with DataType object.
    signature.return_type = orig_type
    assert signature.return_type == orig_type


def test_parameters(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    assert len(signature.parameters) == 2

    param_0 = signature.parameters[0]
    assert param_0.ordinal == 0
    assert param_0.size == 4
    assert "byte *" in param_0.data_type.name.casefold()
    assert isinstance(param_0.location, StackLocation)
    assert param_0.location.stack_offset == 0

    param_1 = signature.parameters[1]
    assert param_1.ordinal == 1
    assert param_1.size == 1
    assert param_1.data_type.name in ("byte", "char")
    assert isinstance(param_1.location, StackLocation)
    assert param_1.location.stack_offset == 4


def test_parameter_info_ida(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    param_0 = signature.parameters[0]
    assert param_0.name == "a1"
    assert param_0.data_type.name == "byte *"
    assert param_0.declaration == "_BYTE * a1"


def test_applied_function_signature_ida(disassembler):
    """
    Test that new signature has been applied back to disassembly view within IDB.
    """
    address = 0x401000
    ida_typeinf = disassembler._ida_typeinf
    ida_nalt = disassembler._ida_nalt

    tif = ida_typeinf.tinfo_t()
    ida_nalt.get_tinfo(tif, address)
    assert str(tif) == ""

    # The act of getting the function signature will use and apply the signature from the decompiler.
    disassembler.get_function_signature(address)
    tif = ida_typeinf.tinfo_t()
    ida_nalt.get_tinfo(tif, address)
    assert str(tif) == "_BYTE *__cdecl(_BYTE *a1, char a2)"


def test_parameter_info_ghidra(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    param_0 = signature.parameters[0]
    assert param_0.name == "param_1"
    assert param_0.data_type.name == "byte *"
    assert param_0.declaration == "[byte * param_1@Stack[0x4]:4]"


def test_modifying_parameters(disassembler):
    signature = disassembler.get_function_signature(0x401000)

    param_0 = signature.parameters[0]
    assert param_0.ordinal == 0

    orig_name = param_0.name
    assert orig_name in ("a1", "param_1")
    param_0.name = "newname"
    assert param_0.name == "newname"
    assert "newname" in signature.declaration
    param_0.name = orig_name

    assert "byte" in param_0.data_type.name.casefold()

    param_0.data_type = "char *"
    assert param_0.data_type.name.casefold() == "char *"
    assert "char *" in signature.declaration.casefold()

    param_0.data_type = "int"
    assert param_0.data_type.name.casefold() == "int"
    assert param_0.data_type.size == 4
    assert "int" in signature.declaration.casefold()

    param_0.data_type = "qword"
    assert param_0.data_type.name.casefold() == "qword"
    assert param_0.data_type.size == 8
    assert "qword" in signature.declaration.casefold()

    # Test inserting a new parameter in the front and see if our param_0
    # variable updates correctly.
    new_param = signature.insert_parameter(0, "int")
    assert new_param.ordinal == 0
    assert param_0.ordinal == 1

    # Remove parameter to reset state.
    signature.remove_parameter(0)
    assert len(signature.parameters) == 2


def test_replace_parameters(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    orig_parameters = signature.parameters
    assert len(orig_parameters) == 2
    orig_param_info = [(param.name, param.data_type.name) for param in orig_parameters]
    # Must pull before modifying, otherwise we get a NotExistsError when trying to access it.
    orig_data_types = signature.parameter_types

    signature.replace_parameters(["int", "char", "int"])
    parameters = signature.parameters
    assert len(parameters) == 3
    assert [param.data_type.name for param in parameters] == ["int", "char", "int"]

    signature.replace_parameters(orig_data_types)
    parameters = signature.parameters
    assert len(parameters) == 2
    assert [(param.name, param.data_type.name) for param in parameters] == orig_param_info


def test_add_parameter(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    assert len(signature.parameters) == 2
    new_param = signature.add_parameter("int")
    assert len(signature.parameters) == 3
    assert new_param == signature.parameters[2]
    assert new_param.data_type.name == "int"
    assert new_param.data_type.size == 4
    assert "int" in signature.declaration.casefold()

    # Remove parameter to reset state.
    signature.remove_parameter(2)
    assert len(signature.parameters) == 2


def test_remove_parameter(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    orig_declaration = signature.declaration
    removed_param_type = signature.parameters[1].data_type.name
    assert len(signature.parameters) == 2
    with pytest.raises(NotExistError):
        signature.remove_parameter(3)
    signature.remove_parameter(1)
    assert len(signature.parameters) == 1
    assert len(signature.declaration) < len(orig_declaration)

    # Add parameter back to reset state.
    signature.add_parameter(removed_param_type)
    assert len(signature.parameters) == 2


def test_remove_parameter_negative(disassembler):
    """
    Test to ensure negative indexing works.
    """
    signature = disassembler.get_function_signature(0x401000)
    orig_declaration = signature.declaration
    removed_param_type = signature.parameters[1].data_type.name
    assert len(signature.parameters) == 2
    signature.remove_parameter(-1)
    assert len(signature.parameters) == 1
    assert len(signature.declaration) < len(orig_declaration)

    # Add parameter back to reset state.
    signature.add_parameter(removed_param_type)
    assert len(signature.parameters) == 2


def test_insert_parameter(disassembler):
    signature = disassembler.get_function_signature(0x401000)
    assert len(signature.parameters) == 2
    with pytest.raises(ValueError):
        signature.insert_parameter(5, "int")
    new_param = signature.insert_parameter(1, "int")
    assert len(signature.parameters) == 3
    assert new_param == signature.parameters[1]
    assert new_param.data_type.name == "int"
    assert "int" in signature.declaration.casefold()

    # Remove parameter to reset state.
    signature.remove_parameter(1)
    assert len(signature.parameters) == 2


def test_insert_parameter_negative(disassembler):
    """
    Test to ensure negative indexing works.
    """
    signature = disassembler.get_function_signature(0x401000)
    assert len(signature.parameters) == 2
    new_param = signature.insert_parameter(-1, "int")
    assert len(signature.parameters) == 3
    assert new_param == signature.parameters[1]
    assert new_param.data_type.name == "int"
    assert "int" in signature.declaration.casefold()

    # Remove parameter to reset state.
    signature.remove_parameter(1)
    assert len(signature.parameters) == 2




import pathlib
import shutil

import pytest

import dragodis


def test_teleport_ida(disassembler):
    # Test basic
    @disassembler.teleport
    def teleported(addr):
        import idc
        return idc.get_strlit_contents(addr)
    assert teleported(0x40C000) == b"Idmmn!Vnsme "

    # Test disassembler gets converted to local version.
    @disassembler.teleport
    def teleported(dis):
        from dragodis.ida.disassembler import IDALocalDisassembler
        return isinstance(dis, IDALocalDisassembler)
    assert teleported(disassembler)


@pytest.mark.parametrize("backend", ["ida", "ghidra"])
def test_shellcode_x86(shared_datadir, backend):
    input_path = shared_datadir / "strings_x86 .text[00401000,0040102a].bin"

    try:
        with dragodis.open_program(str(input_path), backend, processor=dragodis.PROCESSOR_X86) as dis:
            assert dis.processor_name == "x86"
            insn = dis.get_instruction(0x0d)
            assert insn.mnemonic == "movsx"
    except dragodis.NotInstalledError as e:
        pytest.skip(str(e))


@pytest.mark.parametrize("backend", ["ida", "ghidra"])
def test_shellcode_arm(shared_datadir, backend):
    input_path = shared_datadir / "strings_arm .text[000103fc,0001045b].bin"

    try:
        with dragodis.open_program(str(input_path), backend, processor=dragodis.PROCESSOR_ARM) as dis:
            assert dis.processor_name == "ARM"
            # Need to manually define as code in IDA.
            if backend == "ida":
                dis._ida_ua.create_insn(0x14)
            insn = dis.get_instruction(0x14)
            assert insn.mnemonic == "strb"
    except dragodis.NotInstalledError as e:
        pytest.skip(str(e))



"""
Custom script to run rpyc server.
This was done so we can properly control the shutdown of IDA after the server exits.
"""
import os
import site
import sys


def activate_virtualenv():
    """
    Activates the virtualenv found in the inherited VIRTUAL_ENV environment variable.
    """
    base_path = os.environ.get("VIRTUAL_ENV")
    if not base_path:
        return

    # Add to PATH
    bin_dir = os.path.join(base_path, "Scripts" if sys.platform == "win32" else "bin")
    os.environ["PATH"] = os.pathsep.join([bin_dir] + os.environ.get("PATH", "").split(os.pathsep))

    # Add site-packages dir
    if sys.platform == "win32":
        site_packages = os.path.join(base_path, "Lib", "site-packages")
    else:
        site_packages = os.path.join(base_path, "lib", "python{}.{}".format(*sys.version_info), "site-packages")

    site.addsitedir(site_packages)
    sys.real_prefix = sys.prefix
    sys.prefix = base_path


def main():
    """
    Initialize rpyc connection.
    """
    # Activate virtualenv if user was in one.
    activate_virtualenv()

    # Start RPyC server.
    # Importing idc here so we can still import the server externally.
    import idc

    if not idc.ARGV[1:]:
        raise RuntimeError(f"No connection parameter provided.")

    import rpyc
    from rpyc import SlaveService, OneShotServer
    from rpyc.core.stream import NamedPipeStream

    # Don't compress to improve speed.
    rpyc.core.channel.Channel.COMPRESSION_LEVEL = 0

    if sys.platform == "win32":
        pipe_name = idc.ARGV[1]
        stream = NamedPipeStream.create_server(pipe_name)
        with rpyc.classic.connect_stream(stream) as srv:
            srv.serve_all()
    else:
        socket_path = idc.ARGV[1]
        server = OneShotServer(SlaveService, socket_path=socket_path, auto_register=False)
        server.start()

    idc.qexit(0)


if __name__ == "__main__":
    main()




import dragodis


def test_create_reference(disassembler):
    ref = disassembler.create_reference(0x401014, 0x40100d, dragodis.ReferenceType.code_jump)
    assert ref
    assert ref.from_address == 0x401014
    assert ref.to_address == 0x40100d
    assert ref.type == dragodis.ReferenceType.code_jump

    insn = disassembler.get_instruction(0x40100d)
    assert ref in list(insn.references_to)
    insn = disassembler.get_instruction(0x401014)
    assert ref in list(insn.references_from)
